@@ -68,20 +68,17 @@ such as for extremely large ``npartitions`` or if we find we need to
 increase the sample size for each partition.
 
 """
-from __future__ import absolute_import, division, print_function
-
 import math
 import numpy as np
 import pandas as pd
 from pandas.api.types import is_datetime64tz_dtype
 
-from toolz import merge, merge_sorted, take
+from tlz import merge, merge_sorted, take
 
 from ..utils import random_state_data
 from ..base import tokenize
 from .core import Series
 from .utils import is_categorical_dtype
-from dask.compatibility import zip
 
 
 def sample_percentiles(num_old, num_new, chunk_length, upsample=1.0, random_state=None):
@@ -228,9 +225,10 @@ def create_merge_tree(func, keys, token):
         width = tree_width(prev_width)
         groups = tree_groups(prev_width, width)
         keys = [(token, level, i) for i in range(width)]
-        rv.update(
-            (key, (func, list(take(num, prev_keys)))) for num, key in zip(groups, keys)
-        )
+
+        for num, key in zip(groups, keys):
+            rv[key] = (func, list(take(num, prev_keys)))
+
         prev_width = width
         prev_keys = iter(keys)
         level += 1
@@ -473,7 +471,7 @@ def partition_quantiles(df, npartitions, upsample=1.0, random_state=None):
     name3 = "re-quantiles-3-" + token
     last_dsk = {
         (name3, 0): (
-            pd.Series,
+            pd.Series,  # TODO: Use `type(df._meta)` when cudf adds `tolist()`
             (process_val_weights, merged_key, npartitions, (name0, 0)),
             qs,
             None,
